(defconst silcc-mode-version "SiLCC Mode Version 0.000001"
  "Copyright 2008 INRIA, France.")

(defvar comint-prompt-regexp)

(defgroup silcc nil
  "Major mode for editing SiLCC under Emacs."
  :link '(custom-group-link :tag "Font Lock Faces group" font-lock-faces)
  :group 'languages)

(defcustom silcc-indent-width 3
  "Level of indentation in SiLCC buffers."
  :type 'integer
  :group 'silcc)

(defcustom silcc-indent-tabs-mode nil
  "If t, silcc-mode activates indent-tabs-mode and sets tab-width to
silcc-indent-width."
  :type 'integer
  :group 'silcc)

(defgroup silcc-faces nil
  "SiLCC Font Lock Faces group."
  :group 'silcc)

(defface silcc-font-lock-operator-face
  '((((background light)) (:foreground "brown"))
    (t (:foreground "khaki")))
  "Face description for all operators."
  :group 'silcc-faces)
(defvar silcc-font-lock-operator-face
  'silcc-font-lock-operator-face)

(defface silcc-font-lock-keyword-operator-face
  '((((background light)) (:foreground "brown" :bold t))
     (t (:foreground "khaki" :bold t)))
  "Face description for all operators."
  :group 'silcc-faces)
(defvar silcc-font-lock-keyword-operator-face
  'silcc-font-lock-keyword-operator-face)

(defvar silcc-font-lock-keywords
  '(("\\<\\(done\\|result\\|value\\)\\>"
     0 font-lock-builtin-face)
    ("\\(^\\|[ \t[:alnum:]_]\\)\\(!\\|=\\|<>\\|:=\\|<\\|<=\\|>\\|>=\\|\\+\\|-\\|\\*\\|/\\|\\^\\)\\($\\|[ \t[:alnum:]_]\\)"
     2 silcc-font-lock-operator-face)
    ("\\<\\(or\\|and\\|not\\|mod\\|to\\)\\>"
     0 silcc-font-lock-keyword-operator-face)
    ("\\<\\(package\\|module\\|namespace\\|with\\|exists\\|for\\|forall\\|call\\|false\\|true\\|function\\|for\\|in\\|do\\|if\\|else\\|let\\|this\\)\\>"
     0 font-lock-keyword-face)
    ("\\<function\\>[ \t]*\\([[:alpha:]_][[:alnum:]_]*[ \t]*\\.[ \t]*\\)?\\([[:alpha:]_][[:alnum:]_]*\\)("
     2 font-lock-function-name-face)
    (":\\(\\(\\[|\\|\\[\\| \\|\t\\)*[[:alpha:]_][[:alnum:]_]*\\(]\\||]\\| \\|\t\\)*\\)"
     1 font-lock-type-face)
    ("\\<\\([[:alpha:]_][[:alnum:]_']*\\)\\>[^(]"
     1 font-lock-variable-name-face)
    ("\\<0[xX][[:xdigit:]]+\\>"
     0 font-lock-constant-face)
    ("\\<0[oO][0-7]+\\>"
     0 font-lock-constant-face)
    ("\\<0[bB][01]+\\>"
     0 font-lock-constant-face)
    ("\\<[[:digit:]]+\\(\\.[[:digit:]]+\\)?\\([eE][[:digit:]]+\\)?\\>"
     0 font-lock-constant-face))
  "Font-lock keywords for SiLCC mode.")

(defvar silcc-mode-syntax-table
  (let ((table (make-syntax-table)))
    (modify-syntax-entry ?_ "w" table)
    (modify-syntax-entry ?\\ "\\" table)
    (modify-syntax-entry ?/ ". 14" table)
    (modify-syntax-entry ?* ". 23" table)
    (modify-syntax-entry ?+ "." table)
    (modify-syntax-entry ?- "." table)
    (modify-syntax-entry ?= "." table)
    (modify-syntax-entry ?# "<" table)
    (modify-syntax-entry ?\n ">" table)
    (modify-syntax-entry ?< "." table)
    (modify-syntax-entry ?> "." table)
    (modify-syntax-entry ?\" "\"" table)
    table))

(defun silcc-mode ()
  "Major mode for editing SiLCC code.
Commands:
\\{silcc-mode-map}
Entry to this mode calls the value of `silcc-mode-hook'
if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map silcc-mode-map)
  (set-syntax-table silcc-mode-syntax-table)
  (setq major-mode 'silcc-mode)
  (setq mode-name "SiLCC")
  (silcc-mode-variables)
  (set (make-local-variable 'comment-add) 1)
  ;; font lock
  (setq font-lock-defaults
    '(silcc-font-lock-keywords
      nil nil nil
      beginning-of-line))
  (if (fboundp 'run-mode-hooks)
      (run-mode-hooks 'silcc-mode-hook)
    (run-hooks 'silcc-mode-hook))
  (if silcc-indent-tabs-mode
      (progn
        (setq indent-tabs-mode t)
        (setq tab-width silcc-indent-width))
    (setq indent-with-tabs nil)))

(defvar silcc-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map [(control ?c) (control ?d)] 'silcc-toggle-debug-mode)
    (define-key map [(control ?c) (control ?v)] 'silcc-toggle-verbose-mode)
    (define-key map [(control ?c) (control ?c)] 'silcc-compile)
    (define-key map [(control ?c) (control ?b)] 'silcc-build)
    (define-key map [(control ?c) (control ?m)] 'silcc-toggle-project-main-file)
    map))

(defvar silcc-debug-mode nil
  "If equals to t, next compilations will be made with -g option.")

(defun silcc-toggle-debug-mode ()
  "Toggle the value of silcc-debug-mode."
  (interactive)
  (if silcc-debug-mode
      (silcc-set-debug-mode nil)
    (silcc-set-debug-mode t)))

(defun silcc-set-debug-mode (arg)
  "Set silcc-debug-mode to arg and print a corresponding message."
  (if arg
      (progn
        (setq silcc-debug-mode t)
        (message "SiLCC debug is ON."))
    (setq silcc-debug-mode nil)
    (message "SiLCC debug is OFF.")))

(defvar silcc-verbose-mode t
  "If equals to t, next compilations will be made with -v option.")

(defun silcc-toggle-verbose-mode ()
  "Toggle the value of silcc-verbose-mode."
  (interactive)
  (if silcc-verbose-mode
      (silcc-set-verbose-mode nil)
    (silcc-set-verbose-mode t)))

(defun silcc-set-verbose-mode (arg)
  "Set silcc-verbose-mode to arg and print a corresponding message."
  (if arg
      (progn
        (setq silcc-verbose-mode t)
        (message "SiLCC will be verbose."))
    (setq silcc-verbose-mode nil)
    (message "SiLCC will be quiet.")))

(defvar silcc-project-main-file nil
  "If non nil, compile commands operate on this file regardless
  the current buffer.")

(defun silcc-toggle-project-main-file ()
  "Set silcc-project-main-file to the current buffer if this
buffer was not already the main file.  If the buffer was already
the main file, set silcc-project-main-file to nil."
  (interactive)
  (let ((name buffer-file-name))
    (if (eq buffer-file-name silcc-project-main-file)
        (silcc-set-project-main-file nil)
      (silcc-set-project-main-file buffer-file-name))))

(defun silcc-set-project-main-file (arg)
  "Set silcc-project-main-file to arg and print a corresponding message."
  (setq silcc-project-main-file arg)
  (if arg
      (message (concat "'" arg "' is the current SiLCC project main file."))
    (message "SiLCC project main file is unset.")))

(defun silcc-buffer-module-name ()
  "Return the name of the module visited in the current buffer."
  (file-name-sans-extension (file-relative-name buffer-file-name)))

(defun silcc-compile (&optional all)
  "Compile current SiLCC module.

Prefix ARG means recompile everything."
  (interactive (list (if current-prefix-arg t)))
  (silcc-compile-aux all (concat "-c -C " (silcc-buffer-module-name))))

(defun silcc-build (&optional all)
  "Build project main file if any, or current SiLCC module otherwise.

Prefix ARG means recompile everything."
  (interactive (list (if current-prefix-arg t)))
  (let ((module-name nil))
    (if silcc-project-main-file
        (setq module-name silcc-project-main-file)
      (setq module-name (silcc-buffer-module-name)))
    (silcc-compile-aux all module-name)))

(defun silcc-compile-aux (all arg)
  (let
      ((debug "")
       (verbose "")
       (all-option ""))
    (save-buffer)
    (save-some-buffers (not compilation-ask-about-save) nil)
    (if silcc-debug-mode
        (setq debug "-g "))
    (if silcc-verbose-mode
        (setq verbose "-v "))
    (if all
        (setq all-option "-a "))
    (compilation-start (concat "silcc " debug verbose all-option arg))))

(defun silcc-mode-variables ()
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate (concat "%%\\|$\\|" page-delimiter)) ;'%%..'
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'imenu-generic-expression)
  (setq imenu-generic-expression '((nil "^\\sw+" 0)))
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'silcc-indent-line)
  (make-local-variable 'comment-start)
  (setq comment-start "#")
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "\\(?:#+\\|/\\*+\\)[ \t]*")
  (make-local-variable 'comment-end-skip)
  (setq comment-end-skip "[ \t]*\\(\n\\|\\*+/\\)")
  (make-local-variable 'comment-column)
  (setq comment-column 48))

(defun silcc-indent-line ()
  "Indent current line as SiLCC code."
  (interactive "p")
  (let ((indent (silcc-indent-level))
  (pos (- (point-max) (point))))
    (beginning-of-line)
    (indent-line-to indent)
    (if (> (- (point-max) pos) (point))
  (goto-char (- (point-max) pos)))))

(defun silcc-indent-level ()
  "Compute SiLCC indentation level."
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (cond
     ((looking-at "#") comment-column) ;Small comment starts
     ((bobp) 0)                            ;Beginning of buffer
     (t
       (let ((empty t) ind more less)
         ;; Check current line for closing an unclosed paren or middle ; or -> or =>
         (if (looking-at "[^{[(\n]*\\([]});]\\|->\\|=>\\)")
             (setq less t)
           (setq less nil))
         ;; See previous indentation
         (silcc-previous-code-line)
         (if (bobp)
             (setq pind 0)                ;Beginning of buffer
           (setq pind (current-column)))  ;Beginning of clause
         (skip-chars-backward "^\n")
         (let (ind)
           (cond
            ;; if previous line ends in :- or ::= or { or [ or ( or -> or => then indent more
            ((looking-at "[^\n%]*\\(:-\\|::=\\|{\\|\\[\\|(\\|->\\|=>\\)[[:space:]]*\n")
             (setq ind (+ pind silcc-indent-width)))
            ;; if only ; on previous line, indent more
            ((looking-at "[[:space:]]*;[[:space:]]*")
             (setq ind (+ pind silcc-indent-width)))
            ;; if previous line is the end of a clause
            ((looking-at "[^\n]*\\. *\\(#\\|\n\\)")
             (silcc-previous-code-line)
             ;; if the line before the previous line is the end of another
             ;; clause
             (if (or (bobp) (looking-at "[^\n]*\\. *\\(#\\|\n\\)"))
                 (setq ind pind)
               (setq ind (- pind silcc-indent-width))))
            ;; else indent as the previous line
            (t (setq ind pind)))
           (if less
               (- ind silcc-indent-width)
             ind)))))))

(defun silcc-previous-code-line ()
  "Move the cursor to the previous non-empty and non-commented line."
  (let ((empty t))
    (while empty
      (forward-line -1)
      (beginning-of-line)
      (if (bobp)
          (setq empty nil)
        (skip-chars-forward " \t")
        (if (not (or (looking-at "#[^#]") (looking-at "\n")))
            (setq empty nil))))))

(defconst silcc-single-line-error-chars-regexp
  ".*,[[:space:]].*,[[:space:]][^[:digit:]]+[[:space:]]\\([[:digit:]]+\\)-\\([[:digit:]]+\\):"
  "Regexp matching the char numbers for single-line errors.")

(defconst silcc-multi-line-error-chars-regexp
  ".*,[[:space:]].*,[[:space:]][^[:digit:]]+[[:space:]]\\([[:digit:]]+\\)[[:space:]][^[:digit:]]+[[:space:]]\\([[:digit:]]+\\),[[:space:]][^[:digit:]]+[[:space:]]\\([[:digit:]]+\\):"
  "Regexp matching the char numbers for errors which span over several lines.")

(defadvice next-error (after silcc-next-error activate)
 "Read the extra positional information provided by SiLCC.

Puts the point and the mark exactly around the erroneous program
fragment. The erroneous fragment is also temporarily highlighted if
possible."
 (if (eq major-mode 'silcc-mode)
     (let ((beg nil) (end nil) (end-line nil))
       (save-excursion
         (set-buffer "*compilation*")
         (save-excursion
           (goto-char (window-point (get-buffer-window (current-buffer) t)))
           (cond
            ((looking-at silcc-single-line-error-chars-regexp)
             (setq beg (- (string-to-number (match-string 1)) 1)
                   end (- (string-to-number (match-string 2)) 1)))
            ((looking-at silcc-multi-line-error-chars-regexp)
             (setq beg (- (string-to-number (match-string 1)) 1)
                   end-line (string-to-number (match-string 2))
                   end (- (string-to-number (match-string 3)) 1))))))
       (if beg
           (progn
             (beginning-of-line)
             (setq beg (+ (point) beg))
             (if end-line
                 (progn
                   (goto-line end-line)
                   (beginning-of-line)))
             (setq end (+ (point) end))
             (goto-char beg) (push-mark end t t))))))
