open Util

open Format

type arg_option =
  | Zero of (unit -> unit)
  | One of (string -> unit)
  | Optional of (string option -> unit)
  | Tail of (string list -> unit)

let options = ref []

let short_options = Char.Hashtbl.create 17

let long_options = String.Hashtbl.create 17

let register ?short ?long ~help option =
  begin
    match short with
      | None -> ()
      | Some c -> Char.Hashtbl.add short_options c option
  end;
  begin
    match long with
      | None -> ()
      | Some s -> String.Hashtbl.add long_options s option
  end;
  options := (short, long, help) :: !options

let show_help fmt =
  let show option =
    match option with
      | None, None, _ -> ()
      | Some short, None, help ->
          fprintf fmt "@[  @[-%c@]@ \t@[%s@]@]@." short help
      | None, Some long, help ->
          fprintf fmt "@[  @[--%s@]@ \t@[%s@]@]@." long help
      | Some short, Some long, help ->
          fprintf fmt "@[  @[-%c@],@ --@[%s@]@ \t@[%s@]@]@." short long help in
  List.iter show !options

let parse_tail add_file n =
  for i = n to Array.length Sys.argv - 1 do
    add_file Sys.argv.(i)
  done

let rec accu_tail i =
  if i < Array.length Sys.argv then
    Sys.argv.(i) :: accu_tail (succ i)
  else
    []

let parse_long_option parse s i =
  let (option_name, option_arg) =
    try
      let eq_index = String.index_from s 2 '=' in
      let option_name = String.sub s 2 (eq_index - 2) in
      let option_arg =
        String.sub s (eq_index + 1) (String.length s - eq_index - 1) in
      (option_name, Some option_arg)
    with Not_found ->
      (String.sub s 2 (String.length s - 2), None) in
  match Option.find (String.Hashtbl.find long_options) option_name with
    | None -> Err.unknown_option s
    | Some (Zero f) ->
        if option_arg <> None then
          Err.option_does_not_expect_argument ("--" ^ option_name);
        f ();
        parse i
    | Some (One f) ->
        begin
          match option_arg with
            | None ->
                if i >= Array.length Sys.argv then
                  Err.missing_option_argument ("--" ^ option_name);
                f (Sys.argv.(i));
                parse (succ i)
            | Some arg ->
                f arg;
                parse i
        end
    | Some (Optional f) ->
        f option_arg;
        parse i
    | Some (Tail f) ->
        if option_arg <> None then
          Err.option_does_not_expect_argument ("--" ^ option_name);
        f (accu_tail i)

let rec parse_short_option parse s i c =
  if c < String.length s then
    match Option.find (Char.Hashtbl.find short_options) s.[c] with
      | None -> Err.unknown_option ("-" ^ String.make 1 s.[c])
      | Some (Zero f) ->
          f ();
          parse_short_option parse s i (succ c)
      | Some (One f) ->
          let c' = succ c in
          if c' < String.length s then
            begin
              f (String.sub s c' (String.length s - c'));
              parse i
            end
          else
            begin
              if i >= Array.length Sys.argv then
                Err.missing_option_argument s;
              f (Sys.argv.(i));
              parse (succ i)
            end 
      | Some (Optional f) ->
          let c' = succ c in
          if c' < String.length s then
            f (Some (String.sub s c' (String.length s - c')))
          else
            f None;
          parse i
      | Some (Tail f) ->
          let c' = succ c in
          if c' < String.length s then
            Err.option_does_not_expect_argument ("-" ^ String.make 1 s.[c]);
          f (accu_tail i)
  else
    parse i

let parse_short_option parse s i =
  parse_short_option parse s i 1

let parse add_file =
  let rec aux i =
    if i < Array.length Sys.argv then
      match Sys.argv.(i) with
        | "--" -> parse_tail add_file (succ i)
        | s when String.prefix_match "--" s ->
            parse_long_option aux s (succ i)
        | s when String.length s >= 2 && s.[0] = '-' ->
            parse_short_option aux s (succ i)
        | filename ->
            add_file filename;
            aux (succ i)
    else () in
  aux 1
