let sgn_tbl_sz = 17

open Util

open Format

open Big_int

module Const = struct
  type t =
    | Int of big_int
    | Float of float
    | String of string

  let format fmt v =
    match v with
      | Int i -> fprintf fmt "%s" (string_of_big_int i)
      | Float f -> fprintf fmt "%f" f
      | String s -> fprintf fmt "\"%s\"" (String.escaped s)

  let equal c0 c1 = c0 = c1
end

module Term = struct
  type 'a t =
    | Const of Const.t
    | Array of 'a array

  let equal equal_items t0 t1 =
    match t0, t1 with
      | Const c0, Const c1 -> Const.equal c0 c1
      | Array a0, Array a1 -> Array.equal equal_items a0 a1
      | _ -> false

  let format format_items fmt v =
    match v with
      | Const c -> Const.format fmt c
      | Array a -> fprintf fmt "[%a]" (Array.format "," format_items) a

  let map f t =
    match t with
      | Const c -> Const c
      | Array a -> Array (Array.map f a)

  let subst subst_items s t =
    match t with
      | Const _ -> t
      | Array a -> Array (Array.map (subst_items s) a)
end

module Value = struct
  type t =
    | Var of Var.t
    | Term of t Term.t

  let rec equal v0 v1 =
    match v0, v1 with
      | Var v0, Var v1 -> Var.equal v0 v1
      | Term t0, Term t1 -> Term.equal equal t0 t1
      | _ -> false

  let rec occur x v =
    match v with
      | Var y -> Var.equal x y
      | Term (Term.Const _) -> false
      | Term (Term.Array v) -> Array.exists (occur x) v

  let compare v1 v2 =
    match v1, v2 with
      | Var v1, Var v2 -> Var.compare v1 v2
      | _ -> Pervasives.compare v1 v2

  let rec format fmt v =
    match v with
      | Var v -> Var.format fmt v
      | Term t -> Term.format format fmt t

  let rec iter_vars f v =
    match v with
      | Var v -> f v
      | Term (Term.Const _) -> ()
      | Term (Term.Array a) -> Array.iter (iter_vars f) a

  let rec fold_vars f v i =
    match v with
      | Var v -> f v i
      | Term (Term.Const _) -> i
      | Term (Term.Array a) -> Array.fold_right (fold_vars f) a i

  let rec subst s v =
    match v with
      | Var x ->
          begin
            try Var.Map.find x s
            with Not_found -> v
          end
      | Term t -> Term (Term.subst subst s t)

  let of_big_int i = Term (Term.Const (Const.Int i))

  let of_int i = of_big_int (big_int_of_int i)

  let to_var pos v =
    match v with
      | Var x -> x
      | _ -> Err.should_be_a_variable pos
end

module Sgn = struct
  module This = struct
    type t = {
      id: int;
      f: string;
      arity: int
    }

    let compare s s' = Pervasives.compare s.id s'.id

    let hash s = s.id

    let equal s s' = s.id = s'.id
  end

  include This

  let tbl = Hashtbl.Poly.create sgn_tbl_sz

  let max_char = 0x10FFFF

  let count = ref (succ max_char)

  let id_tbl = Int.Hashtbl.create sgn_tbl_sz

  let get f arity =
    let u = Utf8.to_int_array f 0 (String.length f) in
    if arity = 0 && Array.length u = 1 then
      { id = u.(0); f = f; arity = 0 }
    else
      let key = (f, arity) in
      try Hashtbl.Poly.find tbl key
      with Not_found ->
        let id = !count in
        count := succ id;
        let s = { id = id; f = f; arity = arity } in
        Int.Hashtbl.add id_tbl id s;
        Hashtbl.Poly.add tbl key s;
        s

  let of_id id =
    if id <= max_char then
      { id = id; f = Utf8.from_int_array [| id |] 0 1; arity = 0 }
    else
      try Int.Hashtbl.find id_tbl id
      with Not_found -> failwith "of_id"

  module Hashtbl = Hashtbl.Make (This)

  module Set = Set.Make (This)

  module Map = Map.Make (This)
end

module Compound = struct
  type 'a t = {
    s: Sgn.t;
    args: 'a array
  }

  let equal f t0 t1 =
    Sgn.equal t0.s t1.s && Array.equal f t0.args t1.args

  let sid t = t.s.Sgn.id

  let f t = t.s.Sgn.f

  let arity t = t.s.Sgn.arity

  let get t i = t.args.(i)

  let get_last t = get t (pred (arity t))

  let map f t = { t with args = Array.map f t.args }

  let iter f t = Array.iter f t.args

  let mk f args = { s = Sgn.get f (Array.length args); args = args }

  let exists f t = Array.exists f t.args

  let format format_arg fmt t =
    fprintf fmt "%s(%a)" (f t) (Array.format "," format_arg) t.args 

  let compare compare_args t1 t2 =
    Pair.compare Sgn.compare (Array.compare compare_args)
      (t1.s, t1.args) (t2.s, t2.args)

  let fold_left f v t = Array.fold_left f v t.args

  let fold_right f t v = Array.fold_right f t.args v
end

module Agent = struct
  type t =
    | Skip
    | Tell of Pos.range * bool * Var.t * Value.t Compound.t
    | Call of Pos.range * Value.t Compound.t
    | Ask of Pos.range * Var.t * Var.t Compound.t * t
    | Exists of Var.t * t
    | Par of t * t

  let rec big_par l =
    match l with
      | [] -> Skip
      | [a] -> a
      | h :: t -> Par (h, big_par t)

  let rec big_exists l a =
    match l with
      | [] -> a
      | h :: t -> Exists (h, big_exists t a)
 
  let rec format fmt a =
    match a with
      | Skip -> fprintf fmt "1"
      | Tell (_, _, m, t) ->
          fprintf fmt "%a.%a" Var.format m (Compound.format Value.format) t
      | Call (_, t) ->
          fprintf fmt "call:%a " (Compound.format Value.format) t
      | Ask (_, m, t, a0) ->
          fprintf fmt "%a.%a{%a}"
           Var.format m (Compound.format Var.format) t format a0
      | Exists (v, a0) ->
          fprintf fmt "exists %a{%a}" Var.format v format a0
      | Par (a0, a1) ->
          fprintf fmt "%a || %a" format a0 format a1
end
